###############################################################################
#                                                                             #
# Copyright (c) 2018 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import numpy as np
from collections import namedtuple

from beat.backend.python.database import View as BaseView

import bob.io.base
import bob.io.image
import bob.db.casme2


#----------------------------------------------------------


class View(BaseView):
    """Outputs:
        - image: "{{ system_user.username }}/array_4d_uint8/1"
        - file_id: "{{ system_user.username }}/uint64/1"
        - client_id: "{{ system_user.username }}/uint64/1"
        - emotion: "{{ system_user.username }}/text/1"

    One "file_id" is associated with a given "image".
    Several "image" are associated with a given "client_id".
    Several "client_id" are associated with a given "emotion".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   client_id                 | |                   client_id                 |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                          emotion                                            |
    -----------------------------------------------------------------------------------------------
    """

    def index(self, root_folder, parameters):
        Entry = namedtuple('Entry', ['emotion', 'client_id', 'file_id', 'frames'])

        # Open the database and load the objects to provide via the outputs
        db = bob.db.casme2.Database()

        objs = sorted(db.objects(protocol=str(parameters['protocol']),
                                 groups=parameters['group']),
                      key=lambda x: (x.emotion, x.client_id, x.id))

        entries = []

        for obj in objs:
            frames = [ str(os.path.join(obj.make_path(), x.filename)).
                         replace('/idiap/resource/database/CASME2/Cropped', root_folder)
                       for x in obj.frames ]

            entries.append(Entry(obj.emotion, obj.client_id, obj.id, frames))

        return entries


    def get(self, output, index):
        obj = self.objs[index]

        if output == 'emotion':
            return {
                'value': obj.emotion
            }

        elif output == 'client_id':
            return {
                'value': np.uint64(obj.client_id)
            }

        elif output == 'file_id':
            return {
                'value': np.uint64(obj.file_id)
            }

        elif output == 'image':
            frame = bob.io.base.load(obj.frames[0])

            data = np.zeros(shape=(len(obj.frames), frame.shape[0], frame.shape[1], frame.shape[2]), dtype="uint8")
            data[0] = frame

            for i in range(1, len(obj.frames)):
                data[i] = bob.io.base.load(obj.frames[i])

            return {
                'value': data
            }


#----------------------------------------------------------


def setup_tests():
    # Install a mock load function for the images
    def mock_load(root_folder):
        return np.ndarray((3, 10, 20), dtype=np.uint8)

    bob.io.base.load = mock_load


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    view = View()
    view.objs = view.index(
        root_folder='',
        parameters=dict(
            protocol='fold_1',
            group='train',
        )
    )
    view.get('emotion', 0)
    view.get('client_id', 0)
    view.get('file_id', 0)
    view.get('image', 0)
